from .pathglob import pathglob_compile


def parse_lines(input):
    parts = input.split(',')

    ranges = list()

    for p in parts:
        r = p.split('-', 1)

        if len(r) == 1:
            n = int(r[0])
            ranges.append((n, n))
        else:
            f, t = int(r[0]), int(r[1])
            ranges.append((f, t))

    def m(line):
        for (start, end) in ranges:
            if line >= start and line <= end:
                return True

        return False

    return m


class PatternSection:
    def __init__(self, pattern, **kw):
        self.pattern = pattern
        self.expect_license_header = kw.get('expect_license_header', None)
        self.custom_license_header_path = kw.get('custom_license_header_path', None)
        self.start_comment = kw.get('start_comment', None)
        self.end_comment = kw.get('end_comment', None)
        self.skip_header_lines = kw.get('skip_header_lines', None)
        self.skip_header_on_stanza = kw.get('skip_header_on_stanza', None)

    @classmethod
    def build(cls, context, pattern, **kw):
        pattern = pathglob_compile(pattern)

        expect_license_header = kw.get('expect_license_header', None)

        if expect_license_header:
            kw['expect_license_header'] = context.load_license_header(
                expect_license_header)

        custom_license_header_path = kw.get('custom_license_header_path', None)

        if custom_license_header_path:
            kw['custom_license_header_path'] = context.load_license_header_path(
                context.absolute_path(custom_license_header_path))

        skip_header_lines = kw.get('skip_header_lines', None)

        if skip_header_lines:
            kw['skip_header_lines'] = parse_lines(skip_header_lines)

        return PatternSection(pattern, **kw)

    @classmethod
    def parse(cls, context, name, section):
        if not name.startswith('pattern:'):
            raise Exception('Expected pattern section')

        _, pattern = name.split(':', 1)

        expect_license_header = section.get('expect_license_header')
        custom_license_header_path = section.get('custom_license_header_path')
        start_comment = section.get('start_comment')
        end_comment = section.get('end_comment')
        skip_header_lines = section.get('skip_header_lines')
        skip_header_on_stanza = section.get('skip_header_on_stanza')

        return cls.build(context, pattern,
            expect_license_header = expect_license_header,
            custom_license_header_path = custom_license_header_path,
            start_comment = start_comment,
            end_comment = end_comment,
            skip_header_lines = skip_header_lines,
            skip_header_on_stanza = skip_header_on_stanza
        )



import difflib
import itertools
import os

from .violation import Violation
from .licenses import load_license_header
from .licenses import load_license_header_path
from .extensions import load_extension
from .utils import strip_lineend

class FileMatchOptions:
    def __init__(self, global_section, relative, path):
        self.global_section = global_section
        self.relative = relative
        self.path = path
        self.expect_license_header = None
        self.custom_license_header_path = None
        self.start_comment = None
        self.end_comment = None
        self.skip_header_lines = None
        self.skip_header_on_stanza = None

    @property
    def kw(self):
        return {
            "entity": self.global_section.entity,
            "year": self.global_section.year
        }

    def load_section(self, section):
        if section.expect_license_header:
            self.expect_license_header = section.expect_license_header

        if section.custom_license_header_path:
            self.custom_license_header_path = section.custom_license_header_path

        if section.start_comment:
            self.start_comment = section.start_comment

        if section.end_comment:
            self.end_comment = section.end_comment

        if section.skip_header_lines:
            self.skip_header_lines = section.skip_header_lines

        if section.skip_header_on_stanza:
            self.skip_header_on_stanza = section.skip_header_on_stanza

    def evaluate(self):
        _, ext = os.path.splitext(self.path)

        ext = load_extension(ext, self.path, self)

        errors = []

        license_header = None

        if self.expect_license_header is not None:
            license_header = self.expect_license_header

        if self.custom_license_header_path is not None:
            license_header = self.custom_license_header_path

        if license_header is not None:
            errors.append(self.check_expect_line_header(
                self.path, ext, license_header))

        return list(itertools.chain(*errors))

    def render_fixed(self, path, ext, range_index, license_header):
        lines = []

        with open(path) as f:
            for line in f:
                lines.append(line)

        rendered = ext.render_header_comment(license_header.render(**self.kw))

        start_index, end_index = range_index

        for line in lines[:start_index]:
            yield line

        for line, header_line in enumerate(rendered):
            if self.skip_header_lines and self.skip_header_lines(line):
                yield lines[line]
            else:
                yield header_line + u'\n'

        for line in lines[end_index:]:
            yield line

    def fix_expect_line_header(self, path, ext, range_index, license_header):
        fixed = list(self.render_fixed(path, ext, range_index, license_header))

        with open(path, 'w') as f:
            for line in fixed:
                f.write(line)

    def build_diff(self, original_file, path, ext, range_index, license_header):
        def f():
            fixed = list(
                self.render_fixed(path, ext, range_index, license_header)
            )

            return difflib.unified_diff(
                original_file, fixed,
                fromfile=path, tofile=path + '.fix'
            )

        return f

    def check_expect_line_header(self, path, ext, license_header):
        """
        Check that the given license header matches.
        """

        expected_lines = ext.render_header_comment(
            license_header.render(**self.kw)
        )

        original_file = list(open(path))
        file_lines = list(map(strip_lineend, original_file))

        # the last line of the header block
        range_index = ext.find_header_range(file_lines)

        if range_index != (0, 0):
            start_index, end_index = range_index
            file_lines = file_lines[start_index:end_index]

            stanza = self.skip_header_on_stanza

            # skip header stanza is defined and matches a header line.
            if stanza and any(stanza in line for line in file_lines):
                return

        for i, (line, expect) in enumerate(zip(file_lines, expected_lines)):
            if self.skip_header_lines and self.skip_header_lines(i):
                continue

            if line != expect:
                yield Violation(
                    path=path,
                    line=i,
                    message="\"{}\" != \"{}\"".format(line, expect),
                    kind="License Header Mismatch",
                    fix=lambda: self.fix_expect_line_header(
                        path, ext, range_index, license_header
                    ),
                    describe_fix=lambda: "Prepend Header",
                    diff=self.build_diff(
                        original_file, path, ext, range_index, license_header
                    )
                )

                break





from .extension import Extension
from ..utils import strip_lineend

class Java(Extension):
    def __init__(self, path, opt):
        self.path = path
        self.opt = opt
        self.start_comment = opt.start_comment if opt.start_comment else "/**"
        self.end_comment = opt.end_comment if opt.end_comment else "**/"

    def find_header_range(self, lines):
        for i, line in enumerate(lines):
            if i == 0:
                if not line.lstrip().startswith(self.start_comment):
                    break

            if line.rstrip().endswith(self.end_comment):
                return (0, i + 1)

            if line.lstrip().startswith('*'):
                continue

        return (0, 0)

    def render_header_comment(self, lines):
        yield self.start_comment

        for line in lines:
            yield u' * ' + strip_lineend(line)

        yield u' ' + self.end_comment



from .policy import Policy

from ..pattern_section import PatternSection


class Spotify10(Policy):
    name = "Spotify 1.0"

    def __init__(self, options):
        pass

    def apply(self, context, global_section, patterns):
        global_section.entity = 'Spotify AB'
        global_section.expect_license = 'Apache 2.0'
        global_section.auto_year = True

        patterns.append(PatternSection.build(
            context, '/**/*.py',
            expect_license_header = 'Apache 2.0'
        ))

        patterns.append(PatternSection.build(
            context, '*.java',
            start_comment = '/*',
            end_comment = '*/'
        ))

        patterns.append(PatternSection.build(
            context, '/**/src/main/**/*.java',
            expect_license_header = 'Apache 2.0'
        ))



from .java import Java
from .python import Python

EXTENSIONS = {}
EXTENSIONS['.py'] = Python
EXTENSIONS['.java'] = Java

def load_extension(ext, path, opt):
    try:
        ext = EXTENSIONS[ext]
    except KeyError:
        ext = Extension

    return ext(path, opt)



class Extension:
    """
    Base class for extension helpers.
    """

    def __init__(self, path, opt):
        pass

    def find_header_range(self, lines):
        """
        Find the range in lines (zero-based) that the header occupies.
        """

        return (0, 0)

    def render_header_comment(self, lines):
        """
        Render the header comment given the lines of a license.
        """

        for line in lines:
            yield line





import pkg_resources

class License:
    def __init__(self, full, header):
        self.full = full
        self.header = header


class LicenseText:
    def __init__(self, lines):
        self.lines = lines

    def render(self, **kw):
        for line in self.lines:
            yield line.format(**kw)


LICENSES = {}

LICENSES['Apache 2.0'] = License('apache_2.0.txt', 'apache_2.0_header.txt')


def read_license(path):
    content = pkg_resources.resource_string(__name__, path).decode('utf-8')
    content = content.split(u'\n')
    return LicenseText(content)


def load_license_header_path(path):
    content = list(open(path))
    return LicenseText(content)


def load_license_header(key):
    try:
        license = LICENSES[key]
    except KeyError:
        raise Exception('Unsupported license (' + key + ')')

    return read_license(license.header)


def load_license(key):
    try:
        license = LICENSES[key]
    except KeyError:
        raise Exception('Unsupported license (' + key + ')')

    return read_license(license.full)



import datetime

from .licenses import load_license

class GlobalSection:
    def __init__(self):
        self._expect_license = None
        self.year = None
        self.entity = None
        self.base_year = None
        self.auto_year = False

    def set_expect_license(self, name):
        self._expect_license = load_license(name)

    def get_expect_license(self):
        return self._expect_license

    expect_license = property(get_expect_license, set_expect_license)

    def parse_section(self, section):
        expect_license = section.get('expect_license')

        if expect_license:
            self.expect_license = expect_license

        entity = section.get('entity')

        if entity:
            self.entity = entity

        year = section.get('year')

        if year:
            self.year = year

        auto_year = section.getboolean('auto_year')

        if auto_year is not None:
            self.auto_year = auto_year

        base_year = section.get('base_year')

        if base_year:
            self.base_year = base_year

    def verify(self):
        if self.entity is None:
            raise Exception('Missing option \'entity\'')

        if self.year is None:
            if self.auto_year:
                self.year = str(datetime.datetime.now().year)
            else:
                raise Exception('Missing global option \'year\'')

        if self.base_year is not None:
            self.year = "{} - {}".format(self.base_year, self.year)



from .spotify import Spotify10

POLICIES = {}
POLICIES['Spotify 1.0'] = Spotify10

def load_policy(name, options):
    """
    Load and return an instance of the policy with the given name.
    """

    try:
        policy = POLICIES[name]
    except KeyError:
        raise Exception('Unsupported policy (' + name + ')')

    return policy(options)



import re

def pathglob_compile(pattern):
    """
    Implements a strict pattern matching algorithm suitable for file paths.

    Convert the given pattern into a regular expression where:

        `**` - matches any directories.
        `*` - matches anything inside a single directory.
    """

    parts = pattern.split('/')
    results = []

    for p in parts:
        if p == '**':
            results.append('.*')
            continue

        results.append('[^/]+'.join(re.escape(s) for s in p.split('*')))

    # only require a full match if pattern is absolute
    if pattern.startswith('/'):
        return re.compile('^' + '/'.join(results) + '$')

    return re.compile('/'.join(results) + '$')



import os

from .licenses import load_license_header
from .licenses import load_license_header_path

class Context:
    def __init__(self, root):
        self.root = root

    def absolute_path(self, path):
        """
        Get the absolute path as defined by the root of the project.
        """

        if path.startswith('/'):
            return path

        return os.path.join(self.root, path)

    def load_license_header(self, name):
        return load_license_header(name)

    def load_license_header_path(self, path):
        return load_license_header_path(path)



import os
import argparse
import configparser
import sys

from .pathglob import pathglob_compile
from .policies import load_policy

from .context import Context
from .global_section import GlobalSection
from .file_match_options import FileMatchOptions
from .pattern_section import PatternSection

ETC="/etc/fosslint.conf"
DOTFILE=".fosslint"


def setup_parser():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        'root',
        nargs='?',
        metavar="<dir>",
        help="directory to process",
        default=os.getcwd()
    )

    parser.add_argument(
        '--fix',
        help="Fix found problems",
        action="store_const",
        const=True,
        default=False
    )

    parser.add_argument(
        '--yes',
        help="Automatically answer yes when asked to apply a fix",
        action="store_const",
        const=True,
        default=False
    )

    parser.add_argument(
        '-v', '--verbose',
        dest='verbose',
        help="Increase verbosity",
        action="store_const",
        const=True,
        default=False
    )

    return parser


def iterate_files(root):
    queue = [(root, '')]

    while len(queue) > 0:
        (path, rel) = queue.pop()

        for name in os.listdir(path):
            next_path = os.path.join(path, name)
            next_relative = '/'.join([rel, name])

            if os.path.isdir(next_path):
                queue.append((next_path, next_relative))
            else:
                yield (next_path, next_relative)


def wait_for_yes(message):
    while True:
        line = input(message + ' [y/n]? ')
        line = line.strip()

        if line == 'y':
            return True

        if line == 'n':
            return False

        print('Expected \'y\' or \'n\'')


def entry():
    parser = setup_parser()
    ns = parser.parse_args()
    home = os.path.join(os.path.expanduser('~'), DOTFILE)
    configs = [ETC, home, os.path.join(ns.root, DOTFILE)]

    config = configparser.RawConfigParser()

    for c in configs:
        if not os.path.isfile(c):
            continue

        config.read(c)

    # patterns to ignore
    ignored = []
    # patterns to evalute
    patterns = []
    # global configuration
    global_section = GlobalSection()
    # context for local configurations
    context = Context(ns.root)

    for section in config.sections():
        if section.startswith('policy:'):
            _, name = section.split(':', 1)
            options = config.options(section)
            policy = load_policy(name, options)
            print('Applying Policy: ' + policy.name)
            policy.apply(context, global_section, patterns)
            continue

        if section == 'global':
            global_section.parse_section(config[section])
            continue

        if section.startswith('ignore:'):
            _, rest = section.split(':', 1)
            ignored.append(pathglob_compile(rest))
            continue

        p = PatternSection.parse(context, section, config[section])

        if p:
            patterns.append(p)
            continue

        raise Exception('Unsupported section (' + section + ')')

    global_section.verify()

    checks_by_file = dict()

    checks = []

    for (path, relative) in iterate_files(ns.root):
        # is the file ignored
        if any(ign.match(relative) for ign in ignored):
            continue

        for s in patterns:
            if s.pattern.search(relative) is None:
                continue

            try:
                opt = checks_by_file[relative]
            except KeyError:
                opt = checks_by_file[relative] = FileMatchOptions(
                    global_section, relative, path)

            opt.load_section(s)

    for relative, opt in sorted(checks_by_file.items(), key=lambda e: e[0]):
        if ns.verbose:
            print('Checking: ' + relative)

        checks.append((opt, opt.evaluate()))

    if all(len(e) == 0 for p, e in checks):
        print("Performed {} check(s), no issues found :)".format(len(checks)))
        return

    for (opt, errors) in checks:
        for e in errors:
            print("{}:{} - {} ({})".format(e.path, e.line, e.kind, e.message))

            if ns.fix:
                if e.diff:
                    for line in e.diff():
                        sys.stdout.write(line)

                if ns.yes or wait_for_yes(e.describe_fix()):
                    print("Fixing: {}".format(e.path))
                    e.fix()
                else:
                    print("NOT fixing: {}".format(e.path))



class Violation:
    def __init__(self, path, line, **kw):
        self.path = path
        self.line = line
        self.message = kw.pop('message', 'No Error')
        self.kind = kw.pop('kind', 'UNKNOWN')
        self.fix = kw.pop('fix', lambda: ())
        self.describe_fix = kw.pop('describe_fix', lambda: "No Description")
        self.diff = kw.pop('diff', None)



from distutils.core import setup

VERSION = '0.0.7'


setup(
    name='fosslint',
    version=VERSION,
    description="FOSS Policy Checker Tool",
    long_description="""
    fosslint checks the structure of free software projects so that they follow
    a given policy.
    """,
    author='John-John Tedro',
    author_email='udoprog@tedro.se',
    url='http://github.com/udoprog/fosslint',
    license='GPLv3',
    packages=[
        'fosslint',
        'fosslint.licenses',
        'fosslint.policies',
        'fosslint.extensions'
    ],
    scripts=['bin/fosslint'],
    install_requires=[],
    package_data={
        'fosslint.licenses': ['*.txt']
    }
)



from .extension import Extension
from ..utils import strip_lineend

class Python(Extension):
    def __init__(self, path, opt):
        self.path = path
        self.opt = opt

    def find_header_range(self, lines):
        if len(lines) == 0:
            return (0, 0)

        start = 0

        for i, line in enumerate(lines):
            if i == 0 and line.startswith('#!'):
                start = 1
                continue

            if not line.rstrip().startswith('#'):
                return (start, i)

        return (start, i + 1)

    def render_header_comment(self, lines):
        for line in lines:
            yield u"# " + strip_lineend(line)
